from redis import Redis
import time
from langchain_community.document_loaders import AssemblyAIAudioTranscriptLoader
from langchain_community.document_loaders.assemblyai import TranscriptFormat
from pathlib import Path
from langchain_core.documents import Document
from typing import List

class AssemblyTranscription:

    def __init__(self, *, api_key: str, redis: Redis, concurrent_limit: int = 5):
        self.__redis = redis
        self.__limit = concurrent_limit
        self.__api_key = api_key

    @property
    def __limit_key(self) -> str:
        return "AssemblyTranscription:limit"

    def load(self, filepath: str | Path, *, transcript_format=TranscriptFormat.PARAGRAPHS) -> List[Document]:
        # see: https://python.langchain.com/docs/integrations/document_loaders/assemblyai
        loader = AssemblyAIAudioTranscriptLoader(file_path=filepath,
                                                 transcript_format=transcript_format,
                                                 api_key=self.__api_key)
        return loader.load()


    def __wait_for_allow(self):
        max_seconds = 2 * 60  # wait for maximum 2 minute
        tried = 0
        while True:
            if self.__allow_to_proceed():
                return True
            tried += 1
            if tried > max_seconds:
                raise ValueError(f"Failed to convert to text due to maximum waiting has been reached")
            time.sleep(1)

    def __increment(self):
        self.__redis.incr(self.__limit_key)

    def __decrement(self):
        self.__redis.decr(self.__limit_key)

    def __get_current_processing(self) -> int:
        h = self.__redis.get(self.__limit_key)
        if h is None: return 0
        return int(h)

    def __allow_to_proceed(self):
        current = self.__get_current_processing()
        return current < self.__limit



from .base import Aws
from pathlib import Path
from .schemas import TextractResponse


class AwsTextract(Aws):

    def __init__(self, **kwargs):
        super(AwsTextract, self).__init__(**kwargs)

    def textract(self, *, filepath: str | Path) -> TextractResponse:
        """Perform a Tables + Forms + Queries on any given file

        see boto3.textract.detect_document_text response syntax on
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/textract/client/detect_document_text.html
        """
        # Read document content
        with open(filepath, 'rb') as document:
            bytes_data = bytearray(document.read())

        # Amazon Textract client
        session = self.get_session()
        textract = session.client('textract')

        # Call Amazon Textract
        response = textract.detect_document_text(Document={'Bytes': bytes_data})

        # return a pydantic model
        return TextractResponse(**response)



import json
import uuid
from enum import Enum
from functools import reduce
from typing import List, Optional

from pydantic import BaseModel, computed_field, ConfigDict, Field
from keble_helpers import PydanticModelConfig
from redis import Redis


class ProgressTaskStage(str, Enum):
    PENDING = "PENDING"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


class ProgressReport(BaseModel):
    model_config = PydanticModelConfig.default()
    progress: float
    is_root_success: bool
    success: int
    failure: int
    pending: int
    errors: List[str] = []


class ProgressTask(BaseModel):

    model_config = ConfigDict(**PydanticModelConfig.default_dict(), arbitrary_types_allowed=True)
    stage: ProgressTaskStage = ProgressTaskStage.PENDING
    subtasks: List["ProgressTask"] = []
    error: Optional[str] = None
    redis: Optional[Redis] = None
    key: Optional[str] = None
    model_key: Optional[str] = None
    root: Optional["ProgressTask"] = None
    # if the task is assigned to a task
    # it will be marked as True
    # otherwise, it will be marked as False (by default)
    assigned: bool = False

    @property
    def is_root_success(self) -> bool:
        if self.root is not None:
            return self.root.is_root_success
        # self is root
        return self.stage == ProgressTaskStage.SUCCESS

    @property
    def total_assigned(self) -> int:
        if self.root is not None:
            return self.root.total_assigned
        # is root
        return len([s for s in self.subtasks if s.assigned is True])

    @classmethod
    def get_from_redis(cls, redis: Redis, *, key: str) -> Optional["ProgressTask"]:
        h = redis.get(key)
        if h is None: return None
        return ProgressTask(**json.loads(h))

    @classmethod
    def get_prebuilt_subtasks_model(cls, root: "ProgressTask", redis: Redis, *, model_key: str) -> List["ProgressTask"]:
        """number of subtasks will be stored in redis,
        so each time when a new object is initialized,
         a certain amount of subtasks will be prebuilt into the list"""
        h = redis.get(model_key)
        if h is None: return []
        return [ProgressTask(
            root=root,
        ) for _ in range(int(h))]

    def _refresh_redis(self):
        if self.root is not None:
            self.root._refresh_redis()
        else:
            # now self is root
            assert self.key is not None and self.redis is not None, "[Helpers] You must provide redis and key for progress task to refresh redis cache"
            self.redis.set(
                self.key,
                json.dumps(self.model_dump_circular_reference_safe()),
                ex=48 * 60 * 60 # expire after 48 hours
            )
            if self.model_key is not None:
                # we will update how many subtask it requires to complete
                self.redis.set(
                    self.model_key,
                    self.total_assigned,
                    ex=30 * 24 * 60 * 60  # expire after 30 days
                )

    def new_subtask(self) -> "ProgressTask":
        # Avoid setting root to self, set it only to the rootmost instance
        if self.root is not None:
            self.root.new_subtask()
        else:
            # is root, handle subtask by itself
            # before creating any new subtask
            # we first search unassigned task object
            unassigned_list = [s for s in self.subtasks if not s.assigned]
            if len(unassigned_list) > 0:
                subtask = unassigned_list[0]
            else:
                subtask = ProgressTask(root=self)
                self.subtasks.append(subtask) # append this new subtask to list

            subtask.assigned = True # mark as assigned
            self._refresh_redis()
            return subtask


    def success(self):
        self.stage = ProgressTaskStage.SUCCESS
        self._refresh_redis()

    def failure(self, error: Optional[str] = None):
        self.stage = ProgressTaskStage.FAILURE
        self.error = error
        # all pending subtask need to mark as failure
        for s in self.subtasks:
            if s.stage == ProgressTaskStage.PENDING:
                s.failure()
        self._refresh_redis()

    # @computed_field
    @property
    def progress_report(self) -> ProgressReport:
        subtask_progress = [s.progress_report for s in self.subtasks]
        total = 1 + len(self.subtasks)
        progress_floats = [p.progress * (1 / total) for p in subtask_progress]

        success = sum([p.success for p in subtask_progress]) if len(subtask_progress) > 0 else 0
        failure = sum([p.failure for p in subtask_progress]) if len(subtask_progress) > 0 else 0
        pending = sum([p.pending for p in subtask_progress]) if len(subtask_progress) > 0 else 0
        errors = reduce(lambda a, b: a + b, [p.errors for p in subtask_progress]) if len(subtask_progress) > 0 else []
        if self.stage == ProgressTaskStage.SUCCESS:
            success += 1
            progress_floats.append(1 / total)
        if self.stage == ProgressTaskStage.FAILURE:
            failure += 1
            if self.error is not None:
                errors.append(self.error)
        if self.stage == ProgressTaskStage.PENDING:
            pending += 1



        return ProgressReport(
            progress=sum(progress_floats) if len(progress_floats) > 0 else 0,
            is_root_success=self.is_root_success,
            success=success,
            failure=failure,
            pending=pending,
            errors=errors
        )

    def model_dump_circular_reference_safe(self) -> dict:
        """Custom model dump that excludes circular references."""
        return {
            "stage": self.stage,
            "error": self.error,
            "key": self.key,
            "subtasks": [s.model_dump_circular_reference_safe() for s in self.subtasks]
        }
        # return self.model_dump(exclude={"root", "subtasks", "redis"}, mode="json")



from .aliyun import *
from .assembly import *
from .aws import *
from .common import id_generator, generate_random_string, is_pydantic_field_empty, date_to_datetime, datetime_to_date, \
    hash_string, slice_to_list, bad_utf8_str_encoding, get_first_match, ensure_has_folder, zip_dir, remove_dir, \
    wait_generator_stop, is_mime_audio, is_mime_video, is_mime_image, is_mime_media, is_mime_ms_excel, is_mime_csv, \
    is_mime_prefix_in, ms_excel_mimes, audio_mimes, video_mimes, yield_files, get_files, inline_string
from .enum import Environment
from .pydantic import *
from .multithread import threaded, ThreadController
from .progress import *


import boto3


class Aws:

    def __init__(self, *, access_key: str, secret: str, region: str):
        """

        :param access_key: access key id
        :param secret: access key secret
        """
        self.__access_key = access_key
        self.__secret = secret
        self.__region = region

    @property
    def access_key(self):
        return self.__access_key

    @property
    def secret(self):
        return self.__secret

    def get_session(self):
        return boto3.Session(
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret,
            region_name=self.__region
        )



# from pydantic import AnyHttpUrl as Url
# from pydantic.functional_serializers import PlainSerializer
# from typing import Annotated
#
# AnyHttpUrl = Annotated[
#     Url, PlainSerializer(lambda x: str(x), return_type=str, when_used='always')
# ]



from .base import Aws
from .schemas import TextractResponse, TextractBlock, TextractBlockType, TextractGeometry, TextractQuery, TextractPolygon, TextractRelationship, TextractRelationshipType, TextractEntityTypes, TextractTextTypeEnum, TextractBoundingBox, TextractSelectionStatus, TextractDocumentMetadata
from .textract import AwsTextract